import pytest

import zigpy.device
import zigpy.endpoint
import zigpy.group
import zigpy.types as t
import zigpy.zcl

from .async_mock import AsyncMock, MagicMock, call, sentinel

FIXTURE_GRP_ID = 0x1001
FIXTURE_GRP_NAME = "fixture group"


@pytest.fixture
def endpoint(app_mock):
    ieee = t.EUI64(map(t.uint8_t, [0, 1, 2, 3, 4, 5, 6, 7]))
    dev = zigpy.device.Device(app_mock, ieee, 65535)
    return zigpy.endpoint.Endpoint(dev, 3)


@pytest.fixture
def groups(app_mock):
    groups = zigpy.group.Groups(app_mock)
    groups.listener_event = MagicMock()
    groups.add_group(FIXTURE_GRP_ID, FIXTURE_GRP_NAME, suppress_event=True)
    return groups


@pytest.fixture
def group():
    groups_mock = MagicMock(spec_set=zigpy.group.Groups)
    groups_mock.application.mrequest = AsyncMock()
    return zigpy.group.Group(FIXTURE_GRP_ID, FIXTURE_GRP_NAME, groups_mock)


@pytest.fixture
def group_endpoint(group):
    group.request = AsyncMock()
    return zigpy.group.GroupEndpoint(group)


def test_add_group(groups, monkeypatch):
    monkeypatch.setattr(
        zigpy.group,
        "Group",
        MagicMock(spec_set=zigpy.group.Group, return_value=sentinel.group),
    )
    grp_id, grp_name = 0x1234, "Group Name for 0x1234 group."

    assert grp_id not in groups
    ret = groups.add_group(grp_id, grp_name)
    assert groups.listener_event.call_count == 1
    assert ret is sentinel.group

    groups.listener_event.reset_mock()
    ret = groups.add_group(grp_id, grp_name)
    assert groups.listener_event.call_count == 0
    assert ret is sentinel.group


def test_add_group_no_evt(groups, monkeypatch):
    monkeypatch.setattr(
        zigpy.group,
        "Group",
        MagicMock(spec_set=zigpy.group.Group, return_value=sentinel.group),
    )
    grp_id, grp_name = 0x1234, "Group Name for 0x1234 group."

    assert grp_id not in groups
    ret = groups.add_group(grp_id, grp_name, suppress_event=True)
    assert groups.listener_event.call_count == 0
    assert ret is sentinel.group

    groups.listener_event.reset_mock()
    ret = groups.add_group(grp_id, grp_name)
    assert groups.listener_event.call_count == 0
    assert ret is sentinel.group


def test_pop_group_id(groups, endpoint):
    group = groups[FIXTURE_GRP_ID]
    group.add_member(endpoint)
    group.remove_member = MagicMock(side_effect=group.remove_member)
    groups.listener_event.reset_mock()

    assert FIXTURE_GRP_ID in groups
    grp = groups.pop(FIXTURE_GRP_ID)

    assert isinstance(grp, zigpy.group.Group)
    assert FIXTURE_GRP_ID not in groups
    assert groups.listener_event.call_count == 2
    assert group.remove_member.call_count == 1
    assert group.remove_member.call_args[0][0] is endpoint

    with pytest.raises(KeyError):
        groups.pop(FIXTURE_GRP_ID)


def test_pop_group(groups, endpoint):
    assert FIXTURE_GRP_ID in groups
    group = groups[FIXTURE_GRP_ID]
    group.add_member(endpoint)
    group.remove_member = MagicMock(side_effect=group.remove_member)
    groups.listener_event.reset_mock()

    grp = groups.pop(group)
    assert isinstance(grp, zigpy.group.Group)
    assert FIXTURE_GRP_ID not in groups
    assert groups.listener_event.call_count == 2
    assert group.remove_member.call_count == 1
    assert group.remove_member.call_args[0][0] is endpoint

    with pytest.raises(KeyError):
        groups.pop(grp)


def test_group_add_member(group, endpoint):
    listener = MagicMock()
    group.add_listener(listener)

    assert endpoint.unique_id not in group.members
    assert FIXTURE_GRP_ID not in endpoint.member_of
    group.add_member(endpoint)
    assert endpoint.unique_id in group.members
    assert FIXTURE_GRP_ID in endpoint.member_of
    assert listener.member_added.call_count == 1
    assert listener.member_removed.call_count == 0

    listener.reset_mock()
    group.add_member(endpoint)
    assert listener.member_added.call_count == 0
    assert listener.member_removed.call_count == 0

    group.__repr__()
    assert group.name == FIXTURE_GRP_NAME

    with pytest.raises(ValueError):
        group.add_member(endpoint.endpoint_id)


def test_group_add_member_no_evt(group, endpoint):
    listener = MagicMock()
    group.add_listener(listener)

    assert endpoint.unique_id not in group
    group.add_member(endpoint, suppress_event=True)
    assert endpoint.unique_id in group
    assert FIXTURE_GRP_ID in endpoint.member_of
    assert listener.member_added.call_count == 0
    assert listener.member_removed.call_count == 0


def test_noname_group():
    group = zigpy.group.Group(FIXTURE_GRP_ID)
    assert group.name.startswith("No name group ")


def test_group_remove_member(group, endpoint):
    listener = MagicMock()
    group.add_listener(listener)

    group.add_member(endpoint, suppress_event=True)

    assert endpoint.unique_id in group
    assert FIXTURE_GRP_ID in endpoint.member_of
    group.remove_member(endpoint)
    assert endpoint.unique_id not in group
    assert FIXTURE_GRP_ID not in endpoint.member_of
    assert listener.member_added.call_count == 0
    assert listener.member_removed.call_count == 1


def test_group_magic_methods(group, endpoint):
    group.add_member(endpoint, suppress_event=True)

    assert endpoint.unique_id in group.members
    assert endpoint.unique_id in group
    assert group[endpoint.unique_id] is endpoint


def test_groups_properties(groups: zigpy.group.Groups):
    """Test groups properties."""
    assert groups.application is not None


def test_group_properties(group: zigpy.group.Group):
    """Test group properties."""
    assert group.application is not None
    assert group.groups is not None
    assert isinstance(group.endpoint, zigpy.group.GroupEndpoint)


def test_group_cluster_from_cluster_id():
    """Group cluster by cluster id."""

    cls = zigpy.group.GroupCluster.from_id(MagicMock(), 6)
    assert isinstance(cls, zigpy.zcl.Cluster)

    with pytest.raises(KeyError):
        zigpy.group.GroupCluster.from_id(MagicMock(), 0xFFFF)


def test_group_cluster_from_cluster_name():
    """Group cluster by cluster name."""

    cls = zigpy.group.GroupCluster.from_attr(MagicMock(), "on_off")
    assert isinstance(cls, zigpy.zcl.Cluster)

    with pytest.raises(AttributeError):
        zigpy.group.GroupCluster.from_attr(MagicMock(), "no_such_cluster")


async def test_group_ep_request(group_endpoint):
    on_off = zigpy.group.GroupCluster.from_attr(group_endpoint, "on_off")
    await on_off.on()

    assert group_endpoint.device.request.mock_calls == [
        call(
            260,  # profile
            0x0006,  # cluster
            1,  # sequence
            b"\x01\x01\x01",  # data
        )
    ]


def test_group_ep_reply(group_endpoint):
    group_endpoint.request = MagicMock()
    group_endpoint.reply(
        sentinel.cluster,
        sentinel.seq,
        sentinel.data,
        sentinel.extra_arg,
        extra_kwarg=sentinel.extra_kwarg,
    )
    assert group_endpoint.request.call_count == 1
    assert group_endpoint.request.call_args[0][0] is sentinel.cluster
    assert group_endpoint.request.call_args[0][1] is sentinel.seq
    assert group_endpoint.request.call_args[0][2] is sentinel.data
    assert group_endpoint.request.call_args[0][3] is sentinel.extra_arg
    assert group_endpoint.request.call_args[1]["extra_kwarg"] is sentinel.extra_kwarg


def test_group_ep_by_cluster_id(group_endpoint, monkeypatch):
    clusters = {}
    group_endpoint._clusters = MagicMock(return_value=clusters)
    group_endpoint._clusters.__getitem__.side_effect = clusters.__getitem__
    group_endpoint._clusters.__setitem__.side_effect = clusters.__setitem__

    group_cluster_mock = MagicMock()
    group_cluster_mock.from_id.return_value = sentinel.group_cluster
    monkeypatch.setattr(zigpy.group, "GroupCluster", group_cluster_mock)

    assert len(clusters) == 0
    cluster = group_endpoint[6]
    assert cluster is sentinel.group_cluster
    assert group_cluster_mock.from_id.call_count == 1

    assert len(clusters) == 1
    cluster = group_endpoint[6]
    assert cluster is sentinel.group_cluster
    assert group_cluster_mock.from_id.call_count == 1


def test_group_ep_by_cluster_attr(group_endpoint, monkeypatch):
    cluster_by_attr = {}
    group_endpoint._cluster_by_attr = MagicMock(return_value=cluster_by_attr)
    group_endpoint._cluster_by_attr.__getitem__.side_effect = (
        cluster_by_attr.__getitem__
    )
    group_endpoint._cluster_by_attr.__setitem__.side_effect = (
        cluster_by_attr.__setitem__
    )

    group_cluster_mock = MagicMock()
    group_cluster_mock.from_attr.return_value = sentinel.group_cluster
    monkeypatch.setattr(zigpy.group, "GroupCluster", group_cluster_mock)

    assert len(cluster_by_attr) == 0
    cluster = group_endpoint.on_off
    assert cluster is sentinel.group_cluster
    assert group_cluster_mock.from_attr.call_count == 1

    assert len(cluster_by_attr) == 1
    cluster = group_endpoint.on_off
    assert cluster is sentinel.group_cluster
    assert group_cluster_mock.from_attr.call_count == 1


async def test_group_request(group):
    group.application.send_packet = AsyncMock()
    data = b"\x01\x02\x03\x04\x05"
    res = await group.request(
        sentinel.profile,
        sentinel.cluster,
        sentinel.sequence,
        data,
    )
    assert group.application.send_packet.call_count == 1
    packet = group.application.send_packet.mock_calls[0].args[0]

    assert packet.dst == t.AddrModeAddress(
        addr_mode=t.AddrMode.Group, address=group.group_id
    )
    assert packet.profile_id is sentinel.profile
    assert packet.cluster_id is sentinel.cluster
    assert packet.tsn is sentinel.sequence
    assert packet.data.serialize() == data

    assert res.status is zigpy.zcl.foundation.Status.SUCCESS
    assert res.command_id == data[2]


def test_update_group_membership_remove_member(groups, endpoint):
    """New device is not member of the old groups."""

    groups[FIXTURE_GRP_ID].add_member(endpoint)

    assert endpoint.unique_id in groups[FIXTURE_GRP_ID]
    groups.update_group_membership(endpoint, set())

    assert endpoint.unique_id not in groups[FIXTURE_GRP_ID]


def test_update_group_membership_remove_add(groups, endpoint):
    """New device is not member of the old group, but member of new one."""

    groups[FIXTURE_GRP_ID].add_member(endpoint)

    assert endpoint.unique_id in groups[FIXTURE_GRP_ID]
    new_group_id = 0x1234
    assert new_group_id not in groups
    groups.update_group_membership(endpoint, {new_group_id})
    assert endpoint.unique_id not in groups[FIXTURE_GRP_ID]
    assert new_group_id in groups
    assert endpoint.unique_id in groups[new_group_id]


def test_update_group_membership_add_existing(groups, endpoint):
    """New device is member of new and existing groups."""

    groups[FIXTURE_GRP_ID].add_member(endpoint)

    assert endpoint.unique_id in groups[FIXTURE_GRP_ID]
    new_group_id = 0x1234
    groups.add_group(new_group_id)
    assert new_group_id in groups
    groups.update_group_membership(endpoint, {new_group_id, FIXTURE_GRP_ID})
    assert endpoint.unique_id in groups[FIXTURE_GRP_ID]
    assert new_group_id in groups
    assert endpoint.unique_id in groups[new_group_id]



from __future__ import annotations

import dataclasses
import inspect
import typing
from typing import TYPE_CHECKING, Any, ClassVar, Self, cast

import zigpy.types as t

NoneType = type(None)


# To make sure mypy is aware that `IntStruct` is technically a mixin, we need to
# convince it that it really is an integer at runtime
if typing.TYPE_CHECKING:
    IntMixin = int
else:
    IntMixin = object


class ListSubclass(list):
    # So we can call `setattr()` on it
    pass


class EmptyObject:
    # So we can call `setattr()` on it
    pass


@dataclasses.dataclass(frozen=True)
class StructField:
    name: str | None = None
    type: type[Any] | None = None

    requires: typing.Callable[[Struct], bool] | None = dataclasses.field(
        default=None, repr=False
    )
    optional: bool | None = False
    length: typing.Callable[[Struct], int] | None = dataclasses.field(
        default=None, repr=False
    )
    default: Any = dataclasses.field(default=None, repr=False)

    repr: typing.Callable[[typing.Any], str] | None = dataclasses.field(
        default=repr, repr=False
    )

    def replace(self, **kwargs) -> Self:
        return dataclasses.replace(self, **kwargs)

    def _convert_type(self, value):
        if value is None or isinstance(value, self.type):
            return value

        try:
            return self.type(value)
        except Exception as e:  # noqa: BLE001
            raise ValueError(
                f"Failed to convert {self.name}={value!r} from type"
                f" {type(value)} to {self.type}"
            ) from e


if TYPE_CHECKING:

    class ResolvedStructField(StructField):
        """`StructField` instance with name and type resolved."""

        name: str
        type: type[Any]

    class ResolvedArrayStructField(ResolvedStructField):
        """`StructField` instance with name, type, and length resolved."""

        type: type[t.List[Any]]
        length: typing.Callable[[Struct], int]
else:
    ResolvedStructField = StructField
    ResolvedArrayStructField = StructField


class Struct:
    fields: ClassVar[list[ResolvedStructField]]

    @classmethod
    def _real_cls(cls) -> type:
        # The "Optional" subclass is dynamically created and breaks types.
        # We have to use a little introspection to find our real class.
        return next(c for c in cls.__mro__ if c.__name__ != "Optional")

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        # We generate fields up here to fail early and cache it
        cls.fields = cls._real_cls()._get_fields()
        cls._signature = inspect.Signature(
            parameters=[
                inspect.Parameter(
                    name=f.name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=f.default,
                    annotation=f.type,
                )
                for f in cls.fields
            ]
        )

        # Check to see if the Struct is also an integer
        if next(
            (
                c
                for c in cls.__mro__[1:]
                if issubclass(c, t.FixedIntType) and not issubclass(c, Struct)
            ),
            None,
        ) is not None and not issubclass(cls, IntStruct):
            raise TypeError("Integer structs must be subclasses of `IntStruct`")

        cls._hash = -1
        cls._frozen = False

    def __new__(cls: type[Self], *args, **kwargs) -> Self:
        cls = cls._real_cls()  # noqa: PLW0642

        if len(args) == 1 and isinstance(args[0], cls):
            # Like a copy constructor
            if kwargs:
                raise ValueError(f"Cannot use copy constructor with kwargs: {kwargs!r}")

            kwargs = args[0].as_dict()
            args = ()

        # Pretend our signature is `__new__(cls, p1: t1, p2: t2, ...)`
        bound = cls._signature.bind(*args, **kwargs)
        bound.apply_defaults()

        instance = super().__new__(cls)

        # Set each attributes on the instance
        for name, value in bound.arguments.items():
            field = getattr(cls.fields, name)
            setattr(instance, name, field._convert_type(value))

        return instance

    @classmethod
    def _get_fields(cls) -> list[ResolvedStructField]:
        fields = ListSubclass()

        # We need both to throw type errors in case a field is not annotated
        annotations = typing.get_type_hints(cls._real_cls())

        # Make sure every `StructField` is annotated
        for name in vars(cls._real_cls()):
            value = getattr(cls, name)

            if isinstance(value, StructField) and name not in annotations:
                raise TypeError(
                    f"Field {name!r}={value} must have some annotation."
                    f" Use `None` if it is specified in the `StructField`."
                )

        # XXX: Python doesn't provide a simple way to get all defined attributes *and*
        #      order them with respect to annotation-only fields.
        #      Every struct field must be annotated.
        for name, annotation in annotations.items():
            # Skip ClassVar annotations (e.g. inherited from Struct base class)
            if typing.get_origin(annotation) is ClassVar:
                continue

            field = getattr(cls, name, StructField())

            if not isinstance(field, StructField):
                continue

            field = field.replace(name=name)

            # An annotation of `None` means to use the field's type
            if annotation is not NoneType:
                if field.type is not None and field.type != annotation:
                    raise TypeError(
                        f"Field {name!r} type annotation conflicts with provided type:"
                        f" {annotation} != {field.type}"
                    )

                field = field.replace(type=annotation)
            elif field.type is None:
                raise TypeError(f"Field {name!r} has no type")

            resolved_field = cast(ResolvedStructField, field)

            if resolved_field.length is not None and not issubclass(
                resolved_field.type, t.List
            ):
                raise TypeError(
                    f"Field {name!r} has a length function but is not a list type: "
                    f"{resolved_field.type}"
                )

            fields.append(resolved_field)
            setattr(fields, field.name, resolved_field)

        return fields

    def assigned_fields(
        self, *, strict=False
    ) -> list[tuple[ResolvedStructField, typing.Any]]:
        assigned_fields = ListSubclass()

        for field in self.fields:
            value = getattr(self, field.name)

            # Ignore fields that aren't required
            if field.requires is not None and not field.requires(self):
                continue

            # Missing fields cause an error if strict
            if value is None and not field.optional and field.length is None:
                if strict:
                    raise ValueError(
                        f"Value for field {field.name!r} is required: {self!r}"
                    )
                else:
                    # Python bug, the following `continue` is never covered
                    continue  # pragma: no cover

            assigned_fields.append((field, value))
            setattr(assigned_fields, field.name, (field, value))

        return assigned_fields

    @classmethod
    def from_dict(cls: type[Self], obj: dict[str, typing.Any]) -> Self:
        instance = cls()

        for key, value in obj.items():
            field = getattr(cls.fields, key)

            if issubclass(field.type, Struct):
                setattr(instance, field.name, field.type.from_dict(value))
            else:
                setattr(instance, field.name, value)

        return instance

    def as_dict(
        self, *, skip_missing: bool = False, recursive: bool = False
    ) -> dict[str, typing.Any]:
        d = {}

        for f in self.fields:
            value = getattr(self, f.name)

            if value is None and skip_missing:
                continue
            elif recursive and isinstance(value, Struct):
                d[f.name] = value.as_dict(
                    skip_missing=skip_missing, recursive=recursive
                )
            else:
                d[f.name] = value

        return d

    def as_tuple(self, *, skip_missing: bool = False) -> tuple:
        return tuple(self.as_dict(skip_missing=skip_missing).values())

    def serialize(self) -> bytes:
        chunks = []

        bit_offset = 0
        bitfields = []

        for field, value in self.assigned_fields(strict=True):
            if value is None and field.optional:
                continue

            value = field._convert_type(value)

            # Fields with lengths dependent on other fields need to be validated
            if field.length is not None:
                expected_length = field.length(self)

                if expected_length == 0:
                    if value:
                        raise ValueError(
                            f"Field {field.name!r} expected an empty array,"
                            f" got: {value!r}"
                        )

                    # Special case for a list with no elements
                    value = field._convert_type([])
                elif value is None or len(value) != expected_length:
                    raise ValueError(
                        f"Field {field.name!r} expected an array with length"
                        f" {expected_length}, got: {value!r}"
                    )

            # All integral types are compacted into one chunk, unless they start and end
            # on a byte boundary.
            if issubclass(field.type, t.FixedIntType) and not (
                value._bits % 8 == 0 and bit_offset % 8 == 0
            ):
                bit_offset += value._bits
                bitfields.append(value)

                # Serialize the current segment of bitfields once we reach a boundary
                if bit_offset % 8 == 0:
                    chunks.append(t.Bits.from_bitfields(bitfields).serialize())
                    bitfields = []

                continue
            elif bitfields:
                raise ValueError(
                    f"Segment of bitfields did not terminate on a byte boundary: "
                    f" {bitfields}"
                )

            chunks.append(value.serialize())

        if bitfields:
            raise ValueError(
                f"Trailing segment of bitfields did not terminate on a byte boundary: "
                f" {bitfields}"
            )

        return b"".join(chunks)

    @staticmethod
    def _deserialize_internal(
        fields: list[ResolvedStructField], data: bytes
    ) -> tuple[dict[str, typing.Any], bytes]:
        bit_length = 0
        bitfields = []
        result = {}

        # We need to create a temporary instance to call the field's `requires` method,
        # which expects a struct-like object
        temp_instance = EmptyObject()

        for field in fields:
            setattr(temp_instance, field.name, None)

        for field in fields:
            if (field.requires is not None and not field.requires(temp_instance)) or (
                not data and field.optional
            ):
                continue

            if issubclass(field.type, t.FixedIntType) and not (
                field.type._bits % 8 == 0 and bit_length % 8 == 0
            ):
                bit_length += field.type._bits
                bitfields.append(field)

                if bit_length % 8 == 0:
                    if len(data) < bit_length // 8:
                        raise ValueError(f"Data is too short to contain {bitfields}")

                    bits, _ = t.Bits.deserialize(data[: bit_length // 8])
                    data = data[bit_length // 8 :]

                    for f in bitfields:
                        value, bits = f.type.from_bits(bits)
                        result[f.name] = value
                        setattr(temp_instance, f.name, value)

                    assert not bits

                    bit_length = 0
                    bitfields = []

                continue
            elif bitfields:
                raise ValueError(
                    f"Segment of bitfields did not terminate on a byte boundary: "
                    f" {bitfields}"
                )

            if field.length is not None:
                field = cast(ResolvedArrayStructField, field)

                count = field.length(temp_instance)
                items, data = field.type.deserialize(data, count=count)
                value = field.type(items)
            else:
                value, data = field.type.deserialize(data)

            result[field.name] = value
            setattr(temp_instance, field.name, value)

        if bitfields:
            raise ValueError(
                f"Trailing segment of bitfields did not terminate on a byte boundary: "
                f" {bitfields}"
            )

        return result, data

    @classmethod
    def deserialize(cls: type[Self], data: bytes) -> tuple[Self, bytes]:
        fields, data = cls._deserialize_internal(cls.fields, data)
        return cls(**fields), data

    def replace(self, **kwargs: dict[str, typing.Any]) -> Struct:
        d = self.as_dict().copy()
        d.update(kwargs)

        instance = type(self)(**d)

        if self._frozen:
            instance = instance.freeze()

        return instance

    def __eq__(self, other: object) -> bool:
        if not isinstance(self, type(other)) and not isinstance(other, type(self)):
            return NotImplemented

        return self.as_dict() == other.as_dict()

    def _repr_extra_parts(self) -> list[str]:
        extra_parts = []

        if self._frozen:
            extra_parts.append("frozen")

        return extra_parts

    def __repr__(self) -> str:
        fields = []

        # Assigned fields are displayed as `field=value`
        for f, v in self.assigned_fields():
            fields.append(f"{f.name}={f.repr(v)}")

        cls = type(self)

        # Properties are displayed as `*prop=value`
        for attr in dir(cls):
            cls_attr = getattr(cls, attr)

            if not isinstance(cls_attr, property) or hasattr(Struct, attr):
                continue

            value = getattr(self, attr)

            if value is not None:
                fields.append(f"*{attr}={value!r}")

        extra_parts = self._repr_extra_parts()
        if extra_parts:
            extra = f"<{', '.join(extra_parts)}>"
        else:
            extra = ""

        return f"{type(self).__name__}{extra}({', '.join(fields)})"

    @property
    def is_valid(self) -> bool:
        try:
            self.serialize()
        except ValueError:
            return False
        else:
            return True

    def matches(self, other: Struct) -> bool:
        if not isinstance(self, type(other)) and not isinstance(other, type(self)):
            return False

        for field in self.fields:
            actual = getattr(self, field.name)
            expected = getattr(other, field.name)

            if expected is None:
                continue

            if isinstance(expected, Struct):
                if not actual.matches(expected):
                    return False
            elif actual != expected:
                return False

        return True

    def __setattr__(self, name: str, value: typing.Any) -> None:
        if self._frozen:
            raise AttributeError("Frozen structs are immutable, use `replace` instead")

        return super().__setattr__(name, value)

    def __hash__(self) -> int:
        if self._frozen:
            return self._hash

        # XXX: This implementation is incorrect only for a single case:
        # `isinstance(struct, collections.abc.Hashable)` always returns True
        raise TypeError(f"Unhashable type: {type(self)}")

    def freeze(self) -> Self:
        """Freeze a Struct instance, making it hashable and immutable."""
        if self._frozen:
            return self

        kwargs = {}

        for f in self.fields:
            value = getattr(self, f.name)

            if isinstance(value, Struct):
                value = value.freeze()

            kwargs[f.name] = value

        cls = self._real_cls()
        instance = cls(**kwargs)
        instance._hash = hash((cls, tuple(kwargs.items())))
        instance._frozen = True

        return instance


class IntStruct(Struct, IntMixin):
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        try:
            cls._int_type: type[t.FixedIntType] = next(
                c
                for c in cls.__mro__[1:]
                if issubclass(c, t.FixedIntType) and not issubclass(c, Struct)
            )
        except StopIteration:
            raise TypeError("Integer structs must be an integer subclasses") from None

    def __new__(
        cls: type[Self], *args, _underlying_int: int | None = None, **kwargs
    ) -> Self:
        # Integers are immutable in Python so we need to know, at creation time, what
        # the integer value of this object will be. This means that these structs *must*
        # also be immutable.
        cls = cls._real_cls()  # noqa: PLW0642
        underlying_int = _underlying_int

        if len(args) > 1:
            raise TypeError(f"{cls} takes no positional arguments")

        # Like a copy constructor
        if len(args) == 1:
            if not isinstance(args[0], int) or kwargs:
                raise TypeError(
                    f"{cls} can only be constructed from an integer"
                    f" or with just keyword arguments"
                )

            underlying_int = args[0]

            data = cls._int_type(underlying_int).serialize()
            kwargs, _ = cls._deserialize_internal(cls.fields, data)
            args = ()

        if underlying_int is None:
            # To compute the underlying integer, we create a temp instance and serialize
            temp_instance = super(Struct, cls).__new__(cls, 0)

            # Set the correct attributes on the instance so we can serialize
            bound = cls._signature.bind(*args, **kwargs)
            bound.apply_defaults()

            for name, value in bound.arguments.items():
                field = getattr(cls.fields, name)
                setattr(temp_instance, name, value)

            # Finally, serialize
            underlying_int, rest = cls._int_type.deserialize(temp_instance.serialize())
            assert not rest

            # Pretend we were called with the correct kwargs
            args = ()
            kwargs = temp_instance.as_dict()

        bound = cls._signature.bind(*args, **kwargs)
        bound.apply_defaults()

        instance = super(Struct, cls).__new__(cls, underlying_int)

        # Set attributes on the final instance
        for name, value in bound.arguments.items():
            field = getattr(cls.fields, name)
            setattr(instance, name, field._convert_type(value))

        # Freeze it
        instance._frozen = True

        return instance

    __hash__ = int.__hash__

    def _repr_extra_parts(self) -> list[str]:
        # We override this method to omit the unnecessary `<frozen>`
        return [f"{self._int_type(int(self))._hex_repr()}"]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, int):
            raise NotImplementedError

        return int(self) == int(other)

    @classmethod
    def deserialize(cls: type[Self], data: bytes) -> tuple[Self, bytes]:
        fields, remaining = cls._deserialize_internal(cls.fields, data)
        underlying_int, _ = cls._int_type.deserialize(
            data[: len(data) - len(remaining)]
        )

        # We overload deserialization to avoid an unnecessary serialize-deserialize
        # during `cls.__new__` to compute the underlying integer, since we have all the
        # data here already
        return cls(_underlying_int=underlying_int, **fields), remaining



from __future__ import annotations

import asyncio
from collections.abc import Callable
import errno
import logging
import pathlib
from typing import Any, Literal, cast

from serialx import (
    SerialTransport,
    create_serial_connection as serialx_create_serial_connection,
)

from zigpy.typing import UNDEFINED, UndefinedType

LOGGER = logging.getLogger(__name__)


class SerialProtocol(asyncio.Protocol):
    """Base class for packet-parsing serial protocol implementations."""

    def __init__(self) -> None:
        self._buffer = bytearray()
        self._transport: SerialTransport | None = None

        self._connected_event = asyncio.Event()
        self._disconnected_event = asyncio.Event()
        self._disconnected_event.set()

    async def wait_until_connected(self) -> None:
        """Wait for the protocol's transport to be connected."""
        await self._connected_event.wait()

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        LOGGER.debug("Connection made: %s", transport)

        self._transport = cast(SerialTransport, transport)
        self._disconnected_event.clear()
        self._connected_event.set()

    def connection_lost(self, exc: BaseException | None) -> None:
        LOGGER.debug("Connection lost: %r", exc)
        self._connected_event.clear()
        self._disconnected_event.set()
        self._transport = None

    def data_received(self, data: bytes) -> None:
        self._buffer += data

    def close(self) -> None:
        self._buffer.clear()

        if self._transport is not None:
            self._transport.close()

    async def wait_until_closed(self) -> None:
        LOGGER.debug("Waiting for serial port to close")
        await self._disconnected_event.wait()

    async def disconnect(self) -> None:
        self.close()
        await self.wait_until_closed()


async def create_serial_connection(
    loop: asyncio.AbstractEventLoop,
    protocol_factory: Callable[[], asyncio.Protocol],
    url: pathlib.Path | str,
    *,
    baudrate: int = 115200,  # We default to 115200 instead of 9600
    xonxoff: bool | UndefinedType = UNDEFINED,
    rtscts: bool | UndefinedType = UNDEFINED,
    flow_control: Literal["hardware", "software"] | None | UndefinedType = UNDEFINED,
    **kwargs: Any,
) -> tuple[asyncio.Transport, asyncio.Protocol]:
    """Wrapper for serialx that provides simplified flow control kwargs."""

    if flow_control is not UNDEFINED:
        xonxoff = flow_control == "software"
        rtscts = flow_control == "hardware"

    if xonxoff is UNDEFINED:
        xonxoff = False

    if rtscts is UNDEFINED:
        rtscts = False

    LOGGER.debug(
        "Opening a serial connection to %r (baudrate=%s, xonxoff=%s, rtscts=%s)",
        url,
        baudrate,
        xonxoff,
        rtscts,
    )

    url = str(url)

    try:
        transport, protocol = await serialx_create_serial_connection(
            loop,
            protocol_factory,
            url=url,
            baudrate=baudrate,
            xonxoff=xonxoff,
            rtscts=rtscts,
            **kwargs,
        )
    except OSError as exc:
        if exc.errno == errno.EBUSY:
            # Re-raise a more useful exception
            raise PermissionError(
                "The serial port is locked by another application"
            ) from exc

        raise

    return transport, protocol



from __future__ import annotations

import hashlib
import json
import pathlib
from unittest.mock import Mock

from aioresponses import aioresponses
import attrs
import pytest

from tests.conftest import make_node_desc
from tests.ota.test_ota_metadata import image_with_metadata  # noqa: F401
import zigpy.device
from zigpy.ota import OtaImageWithMetadata, providers
import zigpy.types as t

from .conftest import FILES_DIR


def make_device(
    model: str | None = None,
    manufacturer: str | None = None,
    manufacturer_id: int | None = None,
) -> zigpy.device.Device:
    dev = zigpy.device.Device(
        application=Mock(),
        ieee=t.EUI64.convert("00:11:22:33:44:55:66:77"),
        nwk=0x1234,
    )

    dev.node_desc = make_node_desc()

    if manufacturer_id is not None:
        dev.node_desc.manufacturer_code = manufacturer_id

    if model is not None:
        dev.model = model

    if manufacturer is not None:
        dev.manufacturer = manufacturer

    return dev


@attrs.define(frozen=True, kw_only=True)
class SelfContainedOtaImageMetadata(providers.BaseOtaImageMetadata):
    test_data: bytes

    async def _fetch(self) -> bytes:
        return self.test_data


def _test_z2m_index_entry(obj: dict, meta: providers.BaseOtaImageMetadata) -> bool:
    assert meta.checksum == "sha512:" + obj.pop("sha512")
    assert meta.image_type == obj.pop("imageType")
    assert meta.file_size == obj.pop("fileSize")
    assert meta.file_version == obj.pop("fileVersion")
    assert meta.manufacturer_id == obj.pop("manufacturerCode")
    assert meta.min_current_file_version == obj.pop("minFileVersion", None)
    assert meta.max_current_file_version == obj.pop("maxFileVersion", None)
    assert meta.min_hardware_version == obj.pop("hardwareVersionMin", None)
    assert meta.max_hardware_version == obj.pop("hardwareVersionMax", None)
    assert meta.changelog == obj.pop("releaseNotes", None)

    if "modelId" in obj:
        assert meta.model_names == (obj.pop("modelId"),)
    else:
        assert meta.model_names == ()

    if "manufacturerName" in obj:
        assert meta.manufacturer_names == tuple(obj.pop("manufacturerName"))
    else:
        assert meta.manufacturer_names == ()

    return True


async def test_local_z2m_provider():
    index_json = (FILES_DIR / "z2m_index.json").read_text()
    index_obj = json.loads(index_json)

    provider = providers.LocalZ2MProvider(FILES_DIR / "z2m_index.json")

    # Test equality
    assert provider == providers.LocalZ2MProvider(FILES_DIR / "z2m_index.json")
    assert provider != providers.LocalZ2MProvider(FILES_DIR / "z2m_index2.json")
    assert provider != providers.LocalZigpyProvider(FILES_DIR / "z2m_index.json")

    # Compatible with all devices
    assert provider.compatible_with_device(make_device(manufacturer_id=1234))
    assert provider.compatible_with_device(make_device(manufacturer_id=5678))

    index = await provider.load_index()

    assert len(index) == len(index_obj)

    for obj, meta in zip(index_obj, index, strict=True):
        assert _test_z2m_index_entry(obj, meta)

        if isinstance(meta, providers.RemoteOtaImageMetadata):
            assert meta.url == obj.pop("url")
        elif isinstance(meta, providers.LocalOtaImageMetadata):
            assert meta.path == FILES_DIR / obj.pop("path")
            obj.pop("url")
        else:
            pytest.fail(f"Unexpected metadata type: {meta!r}")

        obj.pop("fileName", None)
        obj.pop("otaHeaderString", None)
        obj.pop("originalUrl", None)

        assert not obj


async def test_remote_z2m_provider():
    index_json = (FILES_DIR / "z2m_index.json").read_text()
    index_obj = json.loads(index_json)

    index_url = "https://raw.githubusercontent.com/Koenkk/zigbee-OTA/master/index.json"
    provider = providers.RemoteZ2MProvider(index_url)

    # Compatible with all devices
    assert provider.compatible_with_device(make_device(manufacturer_id=1234))
    assert provider.compatible_with_device(make_device(manufacturer_id=5678))

    with aioresponses() as mock_http:
        mock_http.get(
            index_url,
            body=index_json,
            content_type="text/plain; charset=utf-8",
        )

        index = await provider.load_index()

    assert len(index) == len(index_obj)

    for obj, meta in zip(index_obj, index, strict=True):
        assert _test_z2m_index_entry(obj, meta)
        assert isinstance(meta, providers.RemoteOtaImageMetadata)
        assert meta.url == obj.pop("url")
        obj.pop("path", None)
        obj.pop("fileName", None)
        obj.pop("otaHeaderString", None)
        obj.pop("originalUrl", None)

        assert not obj


async def test_tradfri_provider_dirigera():
    index_json = (FILES_DIR / "ikea_version_info_dirigera.json").read_text()
    index_obj = json.loads(index_json)

    provider = providers.Tradfri()

    # Compatible only with IKEA devices
    assert provider.compatible_with_device(make_device(manufacturer_id=4476))
    assert not provider.compatible_with_device(make_device(manufacturer_id=4477))

    with aioresponses() as mock_http:
        mock_http.get(
            "https://fw.ota.homesmart.ikea.com/DIRIGERA/version_info.json",
            headers={"Location": "https://fw.ota.homesmart.ikea.com/check/update/prod"},
            status=302,
        )

        mock_http.get(
            "https://fw.ota.homesmart.ikea.com/check/update/prod",
            body=index_json,
            content_type="application/json",
        )

        index = await provider.load_index()

    # The provider will not allow itself to be loaded a second time this quickly
    with aioresponses() as mock_http:
        assert (await provider.load_index()) is None
        mock_http.assert_not_called()

    # Skip the gateway firmware
    filtered_version_info_obj = [
        obj
        for obj in index_obj
        if obj["fw_type"] == 2 and obj["fw_image_type"] not in (8710, 8704)
    ]
    assert len(index) == len(index_obj) - 3 == len(filtered_version_info_obj)

    for obj, meta in zip(filtered_version_info_obj, index, strict=True):
        assert isinstance(meta, providers.RemoteOtaImageMetadata)
        assert meta.file_version == int(
            obj["fw_binary_url"].split("_v", 1)[1].split("_", 1)[0]
        )
        assert meta.image_type == obj.pop("fw_image_type")
        assert meta.checksum == "sha3-256:" + obj.pop("fw_sha3_256")
        assert meta.url == obj.pop("fw_binary_url")
        assert meta.manufacturer_id == providers.Tradfri.MANUFACTURER_IDS[0] == 4476

        obj.pop("fw_type")
        assert not obj

    meta = index[0]
    assert meta.image_type == 10242

    ota_contents = (
        FILES_DIR
        / "external/dl/ikea/mgm210l-light-cws-cv-rgbw_release_prod_v268572245_3ae78af7-14fd-44df-bca2-6d366f2e9d02.ota"
    ).read_bytes()

    with aioresponses() as mock_http:
        mock_http.get(
            meta.url,
            body=ota_contents,
            content_type="binary/octet-stream",
        )

        img = await meta.fetch()

    assert img.serialize() == ota_contents


@pytest.mark.parametrize(
    ("index_url", "index_file"),
    [
        (
            "http://fw.ota.homesmart.ikea.net/feed/version_info.json",
            "ikea_version_info_old.json",
        ),
        (
            "http://fw.test.ota.homesmart.ikea.net/feed/version_info.json",
            "ikea_version_info_old_test.json",
        ),
    ],
)
async def test_tradfri_provider_old(index_url: str, index_file: str) -> None:
    index_json = (FILES_DIR / index_file).read_text()
    index_obj = json.loads(index_json)

    provider = providers.Tradfri(index_url)

    # Compatible only with IKEA devices
    assert provider.compatible_with_device(make_device(manufacturer_id=4476))
    assert not provider.compatible_with_device(make_device(manufacturer_id=4477))

    with aioresponses() as mock_http:
        mock_http.get(index_url, body=index_json, content_type="application/json")

        index = await provider.load_index()

    # The provider will not allow itself to be loaded a second time this quickly
    with aioresponses() as mock_http:
        assert (await provider.load_index()) is None
        mock_http.assert_not_called()

    # Skip the gateway firmware
    filtered_version_info_obj = [
        obj
        for obj in index_obj
        if obj["fw_type"] == 2 and obj["fw_image_type"] not in (8710, 8704)
    ]
    assert index
    assert len(index) == len(filtered_version_info_obj)

    for obj, meta in zip(filtered_version_info_obj, index, strict=True):
        assert isinstance(meta, providers.RemoteOtaImageMetadata)
        assert meta.file_version == (
            (obj.pop("fw_file_version_MSB") << 16)
            | (obj.pop("fw_file_version_LSB") << 0)
        )
        assert meta.manufacturer_id == obj.pop("fw_manufacturer_id")
        assert meta.image_type == obj.pop("fw_image_type")
        assert meta.file_size == obj.pop("fw_filesize")
        assert meta.url == obj.pop("fw_binary_url").replace("http://", "https://", 1)

        obj.pop("fw_type")
        assert not obj

    # Pick one of the images common to both feeds
    meta = next(m for m in index if "TRADFRI-motion-sensor-2-" in m.url)
    assert meta.image_type == 4552

    ota_contents = (
        FILES_DIR
        / "external/dl/ikea/10039874-1.0-TRADFRI-motion-sensor-2-2.0.022.ota.ota.signed"
    ).read_bytes()

    with aioresponses() as mock_http:
        mock_http.get(
            meta.url,
            body=ota_contents,
            content_type="binary/octet-stream",
        )

        img = await meta.fetch()

    assert img.serialize() in ota_contents


async def test_tradfri_provider_bad_image() -> None:
    index_json = (FILES_DIR / "ikea_version_info_old.json").read_text()
    provider = providers.Tradfri(
        "http://fw.ota.homesmart.ikea.net/feed/version_info.json"
    )

    with aioresponses() as mock_http:
        mock_http.get(
            "http://fw.ota.homesmart.ikea.net/feed/version_info.json",
            body=index_json,
            content_type="application/json",
        )

        index = await provider.load_index()

    assert index is not None
    meta = next(m for m in index if "TRADFRI-motion-sensor-2-" in m.url)
    assert meta.image_type == 4552

    ota_contents = (
        FILES_DIR
        / "external/dl/ikea/10039874-1.0-TRADFRI-motion-sensor-2-2.0.022.ota.ota.signed"
    ).read_bytes()

    # Flip a bit
    with aioresponses() as mock_http:
        flipped_contents = bytearray(ota_contents)
        flipped_contents[50000] ^= 0b00010000

        mock_http.get(
            meta.url,
            body=bytes(flipped_contents),
            content_type="binary/octet-stream",
        )

        with pytest.raises(ValueError, match="Block 3 has invalid checksum"):
            await meta.fetch()

    # Mess with the header
    with aioresponses() as mock_http:
        bad_contents = bytearray(ota_contents)
        bad_contents[0:4] = b"<htm"

        mock_http.get(
            meta.url,
            body=bytes(bad_contents),
            content_type="binary/octet-stream",
        )

        with pytest.raises(ValueError, match="Invalid signed container"):
            await meta.fetch()


async def test_tradfri_provider_invalid_json():
    index_json = (FILES_DIR / "ikea_version_info_dirigera.json").read_text()
    index_obj = [
        *json.loads(index_json),
        {
            "fw_image_type": 10242,
            "fw_type": 2,
            "fw_sha3_256": "e68e61bd57291e0b6358242e72ee2dfe098cb8b769f572b5b8f8e7a34dbcfaca",
            # We extract the version from the URL. Here, it is invalid.
            "fw_binary_url": "https://fw.ota.homesmart.ikea.com/files/bad.ota",
        },
    ]

    provider = providers.Tradfri()

    with aioresponses() as mock_http:
        mock_http.get(
            "https://fw.ota.homesmart.ikea.com/DIRIGERA/version_info.json",
            body=json.dumps(index_obj),
            content_type="application/json",
        )

        index = await provider.load_index()

    assert len(index) == len(index_obj) - 4


async def test_ledvance_provider():
    index_json = (FILES_DIR / "ledvance_firmwares.json").read_text()
    index_obj = json.loads(index_json)

    provider = providers.Ledvance()

    with aioresponses() as mock_http:
        mock_http.get(
            "https://api.update.ledvance.com/v1/zigbee/firmwares",
            body=index_json,
            content_type="application/json; charset=utf-8",
        )

        index = await provider.load_index()

    assert len(index) == len(index_obj["firmwares"])

    for obj, meta in zip(index_obj["firmwares"], index, strict=True):
        assert isinstance(meta, providers.RemoteOtaImageMetadata)
        assert meta.image_type == obj["identity"]["product"]
        assert meta.checksum == "sha256:" + obj.pop("shA256")
        assert meta.release_notes == obj.pop("releaseNotes")
        assert meta.file_size == obj.pop("length")
        assert meta.manufacturer_id == obj["identity"]["company"]
        assert meta.model_names == (obj.pop("productName"),)
        assert meta.url == (
            f"https://api.update.ledvance.com/v1/zigbee/firmwares/download"
            f"?Company={obj['identity'].pop('company')}"
            f"&Product={obj['identity'].pop('product')}"
            f"&Version={obj['identity']['version'].pop('major')}"
            f".{obj['identity']['version'].pop('minor')}"
            f".{obj['identity']['version'].pop('build')}"
            f".{obj['identity']['version'].pop('revision')}"
        )

        assert not obj["identity"].pop("version")
        assert not obj.pop("identity")

        obj.pop("blob")
        obj.pop("extension")
        obj.pop("fullName")
        obj.pop("name")
        obj.pop("released")
        obj.pop("salesRegion")

        assert not obj


async def test_sonoff_provider():
    index_json = (FILES_DIR / "sonoff_upgrade.json").read_text()
    index_obj = json.loads(index_json)

    provider = providers.Sonoff()

    with aioresponses() as mock_http:
        mock_http.get(
            "https://zigbee-ota.sonoff.tech/releases/upgrade.json",
            body=index_json,
        )

        index = await provider.load_index()

    assert len(index) == len(index_obj)

    for obj, meta in zip(index_obj, index, strict=True):
        assert isinstance(meta, providers.RemoteOtaImageMetadata)
        assert meta.url == obj.pop("fw_binary_url")
        assert meta.file_version == obj.pop("fw_file_version")
        assert meta.file_size == obj.pop("fw_filesize")
        assert meta.image_type == obj.pop("fw_image_type")
        assert meta.manufacturer_id == obj.pop("fw_manufacturer_id")
        assert meta.model_names == (obj.pop("model_id"),)
        assert not obj


async def test_remote_zigpy_provider():
    index_json = (FILES_DIR / "remote_index.json").read_text()
    index_obj = json.loads(index_json)

    # A provider with no manufacturer IDs is compatible with all images
    assert providers.RemoteZigpyProvider("foo").compatible_with_device(
        make_device(manufacturer_id=4476)
    )

    # Ours will have a predefined list, however
    provider = providers.RemoteZigpyProvider(
        "https://example.org/fw/index.json", manufacturer_ids=[1, 2, 3]
    )

    with aioresponses() as mock_http:
        mock_http.get(
            "https://example.org/fw/index.json",
            body=index_json,
            content_type="application/json",
        )

        index = await provider.load_index()

    assert len(index) == len(index_obj["firmwares"])

    for obj, meta in zip(index_obj["firmwares"], index, strict=True):
        assert isinstance(meta, providers.RemoteOtaImageMetadata)
        assert meta.url == obj.pop("binary_url")
        assert meta.file_version == obj.pop("file_version")
        assert meta.file_size == obj.pop("file_size")
        assert meta.image_type == obj.pop("image_type")
        assert meta.manufacturer_names == tuple(obj.pop("manufacturer_names"))
        assert meta.model_names == tuple(obj.pop("model_names"))
        assert meta.manufacturer_id == obj.pop("manufacturer_id")
        assert meta.changelog == obj.pop("changelog")
        assert meta.release_notes == obj.pop("release_notes")
        assert meta.release_url == obj.pop("release_url")
        assert meta.checksum == obj.pop("checksum")
        assert meta.min_hardware_version == obj.pop("min_hardware_version")
        assert meta.max_hardware_version == obj.pop("max_hardware_version")
        assert meta.min_current_file_version == obj.pop("min_current_file_version")
        assert meta.max_current_file_version == obj.pop("max_current_file_version")
        assert meta.specificity == obj.pop("specificity")
        assert not obj

    assert provider.manufacturer_ids == (1, 2, 3)


async def test_local_zigpy_provider():
    index_obj = json.loads((FILES_DIR / "local_index.json").read_text())
    provider = providers.LocalZigpyProvider(FILES_DIR / "local_index.json")
    assert str(provider)

    # Test equality
    assert provider == providers.LocalZigpyProvider(FILES_DIR / "local_index.json")
    assert provider != providers.LocalZigpyProvider(FILES_DIR / "local_index2.json")
    assert provider != providers.LocalZ2MProvider(FILES_DIR / "local_index.json")

    index = await provider.load_index()

    assert len(index) == len(index_obj["firmwares"])

    for obj, meta in zip(index_obj["firmwares"], index, strict=True):
        assert isinstance(meta, providers.LocalOtaImageMetadata)
        assert meta.path == pathlib.Path(FILES_DIR / obj.pop("path"))
        assert meta.file_version == obj.pop("file_version")
        assert meta.file_size == obj.pop("file_size")
        assert meta.image_type == obj.pop("image_type")
        assert meta.manufacturer_names == tuple(obj.pop("manufacturer_names"))
        assert meta.model_names == tuple(obj.pop("model_names"))
        assert meta.manufacturer_id == obj.pop("manufacturer_id")
        assert meta.changelog == obj.pop("changelog")
        assert meta.release_notes == obj.pop("release_notes")
        assert meta.release_url == obj.pop("release_url")
        assert meta.checksum == obj.pop("checksum")
        assert meta.min_hardware_version == obj.pop("min_hardware_version")
        assert meta.max_hardware_version == obj.pop("max_hardware_version")
        assert meta.min_current_file_version == obj.pop("min_current_file_version")
        assert meta.max_current_file_version == obj.pop("max_current_file_version")
        assert meta.specificity == obj.pop("specificity")
        assert not obj


async def test_advanced_file_provider(tmp_path: pathlib.Path) -> None:
    files = list((FILES_DIR / "external/dl/local_provider").glob("[!.]*"))
    files.sort(key=lambda f: f.name)

    if len(files) < 2:
        pytest.skip("Not enough files for testing, assuming a provider failed to load")
        return

    (tmp_path / "foo/bar").mkdir(parents=True)
    (tmp_path / "foo/bar" / files[0].name).write_bytes(files[0].read_bytes())
    (tmp_path / "foo" / files[1].name).write_bytes(files[1].read_bytes())
    (tmp_path / "empty").mkdir(parents=True)
    (tmp_path / "bad.ota").write_bytes(b"This is not an OTA file")

    provider = providers.AdvancedFileProvider(tmp_path)

    # Test equality
    assert provider == providers.AdvancedFileProvider(tmp_path)
    assert provider != providers.AdvancedFileProvider(tmp_path / "foo")
    assert provider != providers.LocalZigpyProvider(tmp_path)

    # The provider is compatible with all devices
    assert provider.compatible_with_device(make_device(manufacturer_id=4476))
    assert provider.compatible_with_device(make_device(manufacturer_id=4454))

    index = await provider.load_index()

    assert index is not None
    index.sort(key=lambda m: m.path.name)

    assert len(index) == len(files)

    for path, meta in zip(files, index, strict=True):
        data = path.read_bytes()

        assert isinstance(meta, providers.LocalOtaImageMetadata)
        assert meta.path.name == path.name
        assert meta.checksum == "sha1:" + hashlib.sha1(data).hexdigest()
        assert meta.file_size == len(data)

        fw = await meta.fetch()
        assert fw.serialize() == data


async def test_ota_fetch_size_and_checksum_validation(
    image_with_metadata: OtaImageWithMetadata,
) -> None:
    assert image_with_metadata.firmware is not None

    meta = SelfContainedOtaImageMetadata(
        file_version=image_with_metadata.metadata.file_version,
        checksum=image_with_metadata.metadata.checksum,
        file_size=image_with_metadata.metadata.file_size,
        test_data=image_with_metadata.firmware.serialize(),
    )

    fw = await meta.fetch()
    assert fw == image_with_metadata.firmware
    assert meta.file_size is not None

    with pytest.raises(ValueError):
        await meta.replace(file_size=meta.file_size + 1).fetch()

    assert meta.checksum is not None
    assert not meta.checksum.endswith("c")

    with pytest.raises(ValueError):
        await meta.replace(checksum=meta.checksum[:-1] + "c").fetch()


async def test_zigpy_ota_provider():
    version_json = (FILES_DIR / "zigpy_ota_version_stable.json").read_text()
    version_obj = json.loads(version_json)  # {"schemas": {"zigpy_v2": {...}}} format
    index_json = (FILES_DIR / "zigpy_ota_index.json").read_text()
    index_obj = json.loads(index_json)  # {"firmwares": [...]} format

    # Test with default stable channel
    provider = providers.ZigpyOtaProvider()
    assert provider.channel == "stable"

    # Compatible with all devices
    assert provider.compatible_with_device(make_device(manufacturer_id=4107))
    assert provider.compatible_with_device(make_device(manufacturer_id=9999))

    version_url = (
        "https://raw.githubusercontent.com/zigpy/zigpy-ota/release/version/stable.json"
    )
    index_url = version_obj["schemas"]["zigpy_v2"]["url"]

    with aioresponses() as mock_http:
        mock_http.get(version_url, body=version_json, content_type="application/json")
        mock_http.get(index_url, body=index_json, content_type="application/json")

        index = await provider.load_index()

    assert len(index) == len(index_obj["firmwares"])

    for obj, meta in zip(index_obj["firmwares"], index, strict=True):
        assert isinstance(meta, providers.RemoteOtaImageMetadata)
        assert meta.url == obj.pop("binary_url")
        assert meta.file_version == obj.pop("file_version")
        assert meta.file_size == obj.pop("file_size")
        assert meta.image_type == obj.pop("image_type")
        assert meta.manufacturer_id == obj.pop("manufacturer_id")
        assert meta.checksum == obj.pop("checksum")
        assert meta.manufacturer_names == tuple(obj.pop("manufacturer_names", []))
        assert meta.model_names == tuple(obj.pop("model_names", []))
        assert meta.release_notes == obj.pop("release_notes", None)
        assert meta.min_hardware_version == obj.pop("min_hardware_version", None)
        assert meta.max_hardware_version == obj.pop("max_hardware_version", None)
        assert meta.min_current_file_version == obj.pop(
            "min_current_file_version", None
        )
        assert meta.max_current_file_version == obj.pop(
            "max_current_file_version", None
        )
        assert meta.specificity == obj.pop("specificity", None)
        assert meta.release_url == obj.pop("release_url", None)

        # Pop any remaining fields that are in the JSON but not in metadata
        obj.pop("source_url", None)

        assert not obj
        assert "zigpy-ota provider (stable channel)" in meta.source

    # The provider should not reload immediately due to caching
    with aioresponses() as mock_http:
        cached_index = await provider.load_index()
        assert cached_index is None
        mock_http.assert_not_called()


@pytest.mark.parametrize(
    ("channel", "version_file", "url"),
    [
        ("dev", "zigpy_ota_version_dev.json", None),
        ("beta", "zigpy_ota_version_stable.json", None),
        (
            "custom",
            "zigpy_ota_version_stable.json",
            "https://example.org/custom/version.json",
        ),
    ],
)
async def test_zigpy_ota_provider_channel(
    channel: str, version_file: str, url: str | None
):
    version_json = (FILES_DIR / version_file).read_text()
    version_obj = json.loads(version_json)
    index_json = (FILES_DIR / "zigpy_ota_index.json").read_text()

    provider = providers.ZigpyOtaProvider(channel=channel, url=url)
    assert provider.channel == channel

    version_url = (
        url
        or f"https://raw.githubusercontent.com/zigpy/zigpy-ota/release/version/{channel}.json"
    )
    index_url = version_obj["schemas"]["zigpy_v2"]["url"]

    with aioresponses() as mock_http:
        mock_http.get(version_url, body=version_json, content_type="application/json")
        mock_http.get(index_url, body=index_json, content_type="application/json")

        index = await provider.load_index()

    assert index is not None
    assert len(index) == 2
    assert all(
        f"zigpy-ota provider ({channel} channel)" in meta.source for meta in index
    )


async def test_zigpy_ota_provider_invalid_channel():
    with pytest.raises(ValueError, match="Invalid channel 'invalid'"):
        providers.ZigpyOtaProvider(channel="invalid")


async def test_zigpy_ota_provider_equality():
    provider1 = providers.ZigpyOtaProvider(channel="stable")
    provider2 = providers.ZigpyOtaProvider(channel="stable")
    provider3 = providers.ZigpyOtaProvider(channel="dev")
    provider4 = providers.ZigpyOtaProvider(url="https://example.org/custom.json")

    assert provider1 == provider2
    assert provider1 != provider3
    assert provider1 != provider4
    assert provider1 != providers.RemoteZigpyProvider("https://example.org/custom.json")

    # Test __hash__ (required for combine_concurrent_calls in get_ota_images)
    assert len({provider1, provider2}) == 1
    assert len({provider1, provider3, provider4}) == 3



import asyncio

import pytest

from zigpy import zdo
import zigpy.device
import zigpy.types as t
import zigpy.zdo.types as zdo_types

from .async_mock import AsyncMock, MagicMock, patch, sentinel


def test_commands():
    for cmdid, cmdspec in zdo.types.CLUSTERS.items():
        assert 0 <= cmdid <= 0xFFFF
        assert isinstance(cmdspec, tuple)
        for paramname, paramtype in zip(cmdspec[0], cmdspec[1], strict=True):
            assert isinstance(paramname, str)
            assert hasattr(paramtype, "serialize")
            assert hasattr(paramtype, "deserialize")


@pytest.fixture
def zdo_f(app):
    ieee = t.EUI64(map(t.uint8_t, [0, 1, 2, 3, 4, 5, 6, 7]))
    dev = zigpy.device.Device(app, ieee, 65535)
    dev.request = AsyncMock()

    app.devices[dev.ieee] = dev

    return zdo.ZDO(dev)


def test_deserialize(zdo_f):
    hdr, args = zdo_f.deserialize(2, b"\x01\x02\x03xx")
    assert hdr.tsn == 1
    assert hdr.is_reply is False
    assert args == [0x0302]


def test_deserialize_unknown(zdo_f):
    with pytest.raises(ValueError):
        hdr, args = zdo_f.deserialize(0x0100, b"\x01")


async def test_request(zdo_f):
    with patch.object(
        zdo_f._device, "get_sequence", wraps=zdo_f._device.get_sequence
    ) as get_sequence:
        await zdo_f.request(2, 65535)
        assert zdo_f.device.request.call_count == 1
        assert zdo_f.device.request.mock_calls[0].kwargs["expect_reply"] is True
        assert get_sequence.call_count == 1


async def test_bind(zdo_f):
    cluster = MagicMock()
    cluster.endpoint.endpoint_id = 1
    cluster.cluster_id = 1026
    await zdo_f.bind(cluster)
    assert zdo_f.device.request.call_count == 1
    assert zdo_f.device.request.mock_calls[0].kwargs["cluster"] == 0x0021


async def test_unbind(zdo_f):
    cluster = MagicMock()
    cluster.endpoint.endpoint_id = 1
    cluster.cluster_id = 1026
    await zdo_f.unbind(cluster)
    assert zdo_f.device.request.call_count == 1
    assert zdo_f.device.request.mock_calls[0].kwargs["cluster"] == 0x0022


@pytest.mark.parametrize(
    ("remove_children", "rejoin", "flags"),
    [
        (False, False, 0),
        (False, True, 0x80),
        (True, False, 0x40),
        (True, True, 0xC0),
    ],
)
async def test_leave(zdo_f, remove_children, rejoin, flags):
    """Test ZDO leave request options."""
    with patch.object(zdo_f, "request", AsyncMock()) as req_mock:
        await zdo_f.leave(remove_children, rejoin)
    assert req_mock.await_count == 1
    assert req_mock.await_args[0][0] == 0x0034
    assert req_mock.await_args[0][1] == t.EUI64.convert("07:06:05:04:03:02:01:00")
    assert req_mock.await_args[0][2] == flags


async def test_permit(zdo_f):
    await zdo_f.permit()
    assert zdo_f.device.request.call_count == 1
    assert zdo_f.device.request.mock_calls[0].kwargs["cluster"] == 0x0036


async def test_broadcast(app):
    await zigpy.zdo.broadcast(app, 0x0036, 0, 0, 60, 0)

    assert app.send_packet.call_count == 1

    packet = app.send_packet.mock_calls[0].args[0]
    assert packet.dst.addr_mode == t.AddrMode.Broadcast
    assert packet.cluster_id == 0x0036


def _handle_match_desc(zdo_f, profile):
    zdo_f.reply = AsyncMock()
    hdr = MagicMock()
    hdr.command_id = zdo_types.ZDOCmd.Match_Desc_req
    zdo_f.handle_message(5, 0x0006, hdr, [None, profile, [], []])
    assert zdo_f.reply.call_count == 1


async def test_handle_match_desc_zha(zdo_f):
    _handle_match_desc(zdo_f, 260)
    await asyncio.wait(asyncio.all_tasks(), return_when=asyncio.FIRST_COMPLETED)
    assert zdo_f.reply.await_count == 1
    assert zdo_f.reply.call_args[0][3]


async def test_handle_match_desc_generic(zdo_f):
    _handle_match_desc(zdo_f, 0)
    await asyncio.wait(asyncio.all_tasks(), return_when=asyncio.FIRST_COMPLETED)
    assert zdo_f.reply.await_count == 1
    assert not zdo_f.reply.call_args[0][3]


async def test_handle_nwk_addr(zdo_f):
    ieee = zdo_f._device.application.state.node_info.ieee
    zdo_f.reply = MagicMock()
    hdr = MagicMock()
    hdr.command_id = zdo_types.ZDOCmd.NWK_addr_req
    zdo_f.handle_message(5, 0x0000, hdr, [ieee, 0x00])
    assert zdo_f.reply.call_count == 1


async def test_handle_ieee_addr(zdo_f):
    nwk = zdo_f._device.application.state.node_info.nwk
    zdo_f.reply = MagicMock()
    hdr = MagicMock()
    hdr.command_id = zdo_types.ZDOCmd.IEEE_addr_req
    zdo_f.handle_message(5, 0x0001, hdr, [nwk, 0x00])
    assert zdo_f.reply.call_count == 1


def test_handle_announce(zdo_f):
    dev = zdo_f._device
    listener = MagicMock()
    zdo_f.add_listener(listener)

    dev._application.devices.pop(dev.ieee)
    hdr = MagicMock()
    hdr.command_id = zdo_types.ZDOCmd.Device_annce
    zdo_f.handle_message(5, 0x0013, hdr, [dev.nwk, dev.ieee, 0])

    assert listener.device_announce.call_count == 1
    assert listener.device_announce.call_args[0][0] is dev

    assert listener.zdo_device_annce.call_count == 1
    assert listener.zdo_device_annce.call_args[0][0] is dev
    assert listener.zdo_device_annce.call_args[0][1] is None
    assert listener.zdo_device_annce.call_args[0][2] is hdr
    assert listener.zdo_device_annce.call_args[0][3] == [dev.nwk, dev.ieee, 0]


def test_handle_permit_join(zdo_f):
    listener = MagicMock()
    zdo_f.add_listener(listener)
    hdr = MagicMock()
    hdr.command_id = zdo_types.ZDOCmd.Mgmt_Permit_Joining_req
    zdo_f.handle_message(5, 0x0036, hdr, [100, 1])
    assert listener.permit_duration.call_count == 1


def test_handle_unsupported(zdo_f):
    listener = MagicMock()
    zdo_f.add_listener(listener)
    hdr = MagicMock()
    hdr.command_id = zdo_types.ZDOCmd(0xFFFF)
    assert hdr.command_id not in list(zdo_types.ZDOCmd)
    zdo_f.request = MagicMock()
    zdo_f.reply = MagicMock()
    zdo_f.handle_message(5, 0xFFFF, hdr, [])

    assert listener.zdo_undefined_0xffff.call_count == 1
    assert zdo_f.request.call_count == 0
    assert zdo_f.reply.call_count == 0


def test_device_accessor(zdo_f):
    assert zdo_f.device.nwk == 65535


async def test_reply(zdo_f):
    zdo_f.device.request = AsyncMock()
    await zdo_f.reply(0x0005)
    assert zdo_f.device.request.call_count == 1


def test_get_attr_error(zdo_f):
    with pytest.raises(AttributeError):
        zdo_f.no_such_attribute()


async def test_reply_tsn_override(zdo_f, monkeypatch):
    clusters = MagicMock()
    clusters.__getitem__.return_value = (
        sentinel.param_names,
        sentinel.scheam,
    )
    monkeypatch.setattr(zdo_types, "CLUSTERS", clusters)
    mock_ser = MagicMock()
    mock_ser.return_value = b"\xaa\x55"
    monkeypatch.setattr(t, "serialize", mock_ser)
    await zdo_f.reply(sentinel.cmd, sentinel.arg1, sentinel.arg2)
    seq = zdo_f.device.request.mock_calls[0].kwargs["sequence"]
    data = zdo_f.device.request.mock_calls[0].kwargs["data"]
    assert seq == 1
    assert data[0] == 1
    assert data[1:3] == b"\xaa\x55"

    # override tsn
    tsn = 0x23
    await zdo_f.reply(sentinel.cmd, sentinel.arg1, sentinel.arg2, tsn=tsn)
    seq = zdo_f.device.request.mock_calls[1].kwargs["sequence"]
    data = zdo_f.device.request.mock_calls[1].kwargs["data"]
    assert seq == tsn
    assert data[0] == tsn
    assert data[1:3] == b"\xaa\x55"
